<?php /* Fetch available models */
function askme_fetch_openai_models() {
	$models = get_transient('askme_openai_models');
	if ($models !== false) {
		return $models;
	}

	$chatgpt_api_key = askme_options("chatgpt_api_key");
	if ($chatgpt_api_key != "") {
		$url = 'https://api.openai.com/v1/models';
		$headers = array(
			'Authorization' => 'Bearer ' . $chatgpt_api_key,
		);

		$response = wp_remote_get($url, array(
			'headers' => $headers,
		));

		if (is_wp_error($response)) {
			error_log('OpenAI API request failed: ' . $response->get_error_message());
			return [];
		}

		$response_code = wp_remote_retrieve_response_code($response);
		if ($response_code == 200) {
			$body = wp_remote_retrieve_body($response);
			$result = json_decode($body, true);
			$models_result = isset($result['data']) ? $result['data'] : [];
			$models = array();
			foreach ($models_result as $value) {
				$models[] = array("name" => $value['id'],"checkit" => "no");
			}
			$checked_openai_modules_count = get_option("askme_checked_openai_modules_count");
			$count_models = count($models);
			if ($checked_openai_modules_count != $count_models) {
				update_option("askme_check_openai_models",true);
			}
			$unchecked_openai_modules = get_option("askme_unchecked_openai_modules");
			$unchecked_openai_modules = (is_array($unchecked_openai_modules)?$unchecked_openai_modules:array());
			update_option("askme_unchecked_openai_modules",array_merge($models,$unchecked_openai_modules));
			update_option("askme_unchecked_openai_modules_count",$count_models);
			set_transient('askme_openai_models',$models,WEEK_IN_SECONDS);
			return $models;
		}else {
			error_log('OpenAI API request failed with status code: ' . $response_code);
			return [];
		}
	}else {
		error_log('Please check the API key.');
		return [];
	}
}
/* Check Access to Models */
function askme_check_free_model($model_id) {
	$free_models = array(
		'gpt-3.5-turbo',
		'gpt-3.5-turbo-0301',
		'gpt-3.5-turbo-0613',
		'gpt-3.5-turbo-1106',
		'gpt-3.5-turbo-0125',
		'gpt-3.5-turbo-instruct',
		'gpt-3.5-turbo-instruct-0914'
	);
	return (in_array($model_id,$free_models)?'free':'paid');
}
/* Remove Other to Models */
function askme_remove_unused_model($model_id) {
	$unused_models = array(
		'text-embedding-ada-002',
		'text-embedding-3-large',
		'text-embedding-3-small',
		'tts-1-1106',
		'tts-1',
		'tts-1-hd-1106',
		'tts-1-hd',
		'dall-e-2',
		'dall-e-3',
		'whisper-1'
	);
	return (in_array($model_id,$unused_models)?'unused':'');
}
/* Categorize Models */
function askme_categorize_models() {
	$models = get_transient('askme_chekced_openai_models');
	if ($models !== false) {
		return $models;
	}

	$unchecked_openai_modules = get_option("askme_unchecked_openai_modules");
	$unchecked_openai_modules = (is_array($unchecked_openai_modules)?$unchecked_openai_modules:array());

	$checked_openai_modules = get_option("askme_checked_openai_modules");
	$checked_openai_modules = (is_array($checked_openai_modules)?$checked_openai_modules:array());

	if (empty($unchecked_openai_modules) || empty($checked_openai_modules)) {
		askme_fetch_openai_models();
		$unchecked_openai_modules = get_option("askme_unchecked_openai_modules");
		$unchecked_openai_modules = (is_array($unchecked_openai_modules)?$unchecked_openai_modules:array());

		$checked_openai_modules = get_option("askme_checked_openai_modules");
		$checked_openai_modules = (is_array($checked_openai_modules)?$checked_openai_modules:array());
	}

	if (!empty($unchecked_openai_modules)) {
		// Extract names from both arrays
		$names1 = array_column($unchecked_openai_modules, 'name');
		$names2 = array_column($checked_openai_modules, 'name');

		// Find unique names in each array
		$uniqueNames1 = array_diff($names1, $names2);
		$uniqueNames2 = array_diff($names2, $names1);

		// Filter original arrays based on unique names
		$filteredArray1 = array();
		foreach ($unchecked_openai_modules as $item) {
			if (in_array($item['name'], $uniqueNames1)) {
				$filteredArray1[] = $item;
			}
		}

		$filteredArray2 = array();
		foreach ($checked_openai_modules as $item) {
			if (in_array($item['name'], $uniqueNames2)) {
				$filteredArray2[] = $item;
			}
		}

		// Combine the filtered arrays if needed and get 5 items
		$difference = array_merge($filteredArray1, $filteredArray2);
		$difference = array_slice($difference, 0, 5);

		$models = [];

		foreach ($difference as $key => $model) {
			if (askme_check_free_model($model['name']) == 'free') {
				$models[] = array('name' => $model['name'],'checkit' => 'yes','free' => 'free');
			}else {
				$models[] = array('name' => $model['name'],'checkit' => 'yes','paid' => 'paid');
			}
		}

		$models = array_merge($checked_openai_modules,$models);
		update_option("askme_checked_openai_modules",$models);

		$checked_openai_modules = get_option("askme_checked_openai_modules");
		$checked_openai_modules = (is_array($checked_openai_modules)?$checked_openai_modules:array());

		$count_models = count($checked_openai_modules);
		update_option("askme_checked_openai_modules_count",$count_models);

		$unchecked_openai_modules_count = (int)get_option("askme_unchecked_openai_modules_count");
		$checked_openai_modules_count = (int)get_option("askme_checked_openai_modules_count");

		if ($unchecked_openai_modules_count > 0 && $unchecked_openai_modules_count == $checked_openai_modules_count) {
			set_transient('askme_chekced_openai_models',$models,WEEK_IN_SECONDS);
		}
	}

	return $checked_openai_modules;
}
/* Get response */
function askme_chatgpt_response($question_text = "",$type = "answer",$prompt = "") {
	$chatgpt_api_key = askme_options("chatgpt_api_key");
	if ($chatgpt_api_key != "") {
		$chatgpt_model = askme_options("chatgpt_model");
		$chatgpt_model = ($chatgpt_model != ""?$chatgpt_model:"gpt-3.5-turbo");
		$chatgpt_language = askme_options("chatgpt_language");
		$chatgpt_language = ($chatgpt_language != ""?$chatgpt_language:"English");
		$wordpress_languages = askme_wordpress_languages();
		$chatgpt_language = (isset($wordpress_languages[$chatgpt_language]["name"])?$wordpress_languages[$chatgpt_language]["name"]:"English");
		$chatgpt_prompt = askme_options("chatgpt_prompt");
		$chatgpt_role_content = askme_options("chatgpt_role_content");
		$chatgpt_role_content = ($chatgpt_role_content != ""?$chatgpt_role_content:"You are a helpful assistant.");
		
		if ($prompt == "") {
			$prompt = "Please if you don't know the answer or need more explantion just return with ASKMENOANSWER without send any another word, Answer the following questions in the ".$chatgpt_language." language, Here is the question: ".$question_text.($chatgpt_prompt != ""?" ".$chatgpt_prompt:"").apply_filters("askme_extra_prompt","",$type);
		}else {
			$prompt .= $prompt."Please make all prompt in the ".$chatgpt_language;
		}
		
		$url = 'https://api.openai.com/v1/chat/completions';
		if (strpos($chatgpt_model,'text-embedding-') === 0) {
			$url = 'https://api.openai.com/v1/embeddings';
		}else if (strpos($chatgpt_model,'text-') === 0 || strpos($chatgpt_model,'babbage-') === 0 || strpos($chatgpt_model,'davinci-') === 0) {
			$url = 'https://api.openai.com/v1/completions';
		}else if (strpos($chatgpt_model,'dall-e-') === 0) {
			$url = 'https://api.openai.com/v1/images/generations';
		}else if (strpos($chatgpt_model,'whisper-') === 0) {
			$url = 'https://api.openai.com/v1/audio/transcriptions';
		}else if (strpos($chatgpt_model,'tts-') === 0) {
			$url = 'https://api.openai.com/v1/text-to-speech';
		}

		$data = array(
			'model' => $chatgpt_model,
			'max_tokens' => apply_filters("askme_chatgpt_max_tokens",300)
		);

		if ($url == 'https://api.openai.com/v1/chat/completions') {
			$data = array_merge(array('messages' => array(array('role' => 'system', 'content' => $chatgpt_role_content),array('role' => 'user', 'content' => $prompt))),$data);
		}else if ($url == 'https://api.openai.com/v1/embeddings' || $url == 'https://api.openai.com/v1/text-to-speech') {
			$data = array_merge(array('input' => $prompt),$data);
		}else {
			$data = array_merge(array('prompt' => $prompt),$data);
		}

		$data = apply_filters("askme_chatgpt_response_args",$data,$type);

		$headers = array(
			'Content-Type' => 'application/json',
			'Authorization' => 'Bearer ' . $chatgpt_api_key,
		);

		// Retry logic with exponential backoff
		$max_retries = 5;
		$initial_wait_time = 10; // Start with a longer initial wait time
		$response = false;

		for ($attempt = 0; $attempt < $max_retries; $attempt++) {
			$response = wp_remote_post($url, array(
				'headers' => $headers,
				'body' => json_encode($data),
				'timeout' => 60,
			));

			if (is_wp_error($response)) {
				error_log('OpenAI API request failed: ' . $response->get_error_message());
				return 'Error: Unable to get response from OpenAI API.';
			}

			$response_code = wp_remote_retrieve_response_code($response);
			if ($response_code == 429) {
				// Wait before retrying
				sleep($initial_wait_time);
				$initial_wait_time *= 2; // exponential backoff
				continue; // retry the request
			}

			if ($response_code == 200) {
				$body = wp_remote_retrieve_body($response);
				$result = json_decode($body, true);
				if (isset($result['choices'])) {
					if ($type == "answer") {
						$responses = array();
						foreach ($result['choices'] as $choice) {
							$content = $choice['message']['content'];
							$content = trim($content);
							if ($content != "ASKMENOANSWER" || strpos($content,'ASKMENOANSWER') === false) {
								$content = preg_replace('/(?<!\d\.)\s*\n\s*(?!\d\.)/','</p><p>',$content);
								$content = preg_replace('/```(\w*)/', '<pre>', $content, 1);
	    						$content = preg_replace('/```(?!.*```)/', '</pre>', $content);
	    						$content = preg_replace('/(\d\.\s*[^<]*)\n/','$1<br>',$content);
	    						$content = preg_replace('/\*\*(.*?)\*\*/','<strong>$1</strong>',$content);
	    						$content = preg_replace('/###\s*/','',$content);
							}
							$responses[] = $content;
						}
						return $responses;
					}else {
						return $result['choices'];
					}
				}
			}else {
				error_log('OpenAI API request failed with status code: ' . $response_code);
				return false;
			}
		}
		error_log('Error: Unable to get response from OpenAI API after multiple attempts.');
		return false;
	}else {
		error_log('Please check the API key.');
		return false;
	}
}
/* Get WordPress languages */
function askme_wordpress_languages() {
	$languages = get_transient('askme_wordpress_languages');
	if ($languages !== false) {
		return $languages;
	}

	require_once ABSPATH . 'wp-admin/includes/translation-install.php';
	$available_translations = wp_get_available_translations();
	$languages = array();
	foreach ($available_translations as $lang) {
		$languages[$lang['language']] = array('lang' => $lang['language'],'name' => $lang['english_name']);
	}
	update_option("askme_wordpress_languages",$languages);
	set_transient('askme_wordpress_languages',$languages,MONTH_IN_SECONDS);

	return $languages;
}
/* Check title and content */
function askme_check_question_content($post_title = "",$post_content = "") {
	$question_text = "";
	if (!empty($post_title) && !empty($post_content)) {
		$question_text = "Question Title: " . $post_title . "\n\n" . "Question Content: " . $post_content;
	}else if (!empty($post_title) && empty($post_content)) {
		$question_text = $post_title;
	}else if (empty($post_title) && !empty($post_content)) {
		$question_text = $post_content;
	}
	return $question_text;
}
/* Add answer for questions */
add_action('askme_after_adding_posts','askme_chatgpt_transition_post');
function askme_chatgpt_transition_post($post) {
	if ($post->post_type == ask_questions_type || $post->post_type == ask_asked_questions_type) {
		$activate_ai = askme_options('activate_ai');
		if ($activate_ai == 1) {
			$post_id = $post->ID;
			$post_title = strip_tags($post->post_title);
			$post_content = strip_tags($post->post_content);
			$question_text = askme_check_question_content($post_title,$post_content);
			$answer_it = apply_filters("askme_filter_answer_chatgpt",true,$post);
			if ($answer_it == true) {
				$response_text = askme_chatgpt_response($question_text);
				$mysql_time = current_time('mysql');
				$k = 0;
				if (is_array($response_text) && !empty($response_text)) {
					foreach ($response_text as $response) {
						if ($response != "ASKMENOANSWER" || strpos($response,'ASKMENOANSWER') === false) {
							$k++;
							askme_add_chatgpt_answer($post_id,$response,$mysql_time,$k);
						}
					}
				}else {
					error_log('Could not generate an answer for the question: '.$question_text);
				}
			}
		}
	}
}
function askme_add_chatgpt_answer($post_id,$response_text,$mysql_time,$k) {
	$chatgpt_best_answer = askme_options('chatgpt_best_answer');
	$user_id = askme_options('chatgpt_answer_user');

	$user_id = apply_filters('askme_user_answer_chatgpt',$user_id,$post_id);
	$time = apply_filters('askme_time_answer_chatgpt',$mysql_time);

	$user = get_user_by('id',$user_id);
	if (!$user) {
		$user_query = new WP_User_Query(array('role' => 'Administrator','count_total' => false,'fields' => 'ID'));
		$user_get_results = $user_query->get_results();
		$user_id = $user_get_results[0];
	}

	$comment_data = array(
		'comment_post_ID' => $post_id,
		'comment_author' => $user->display_name,
		'comment_author_email' => $user->user_email,
		'comment_author_url' => $user->user_url,
		'comment_content' => str_ireplace("ASKMENOANSWER","",$response_text),
		'user_id' => $user_id,
		'comment_date' => $time,
		'comment_date_gmt' => get_gmt_from_date($time),
		'comment_approved' => 1,
	);
	
	$chatgpt_answer_type_of_users = askme_options("chatgpt_answer_type_of_users");
	if ($chatgpt_answer_type_of_users == "anonymously") {
		$comment_data["user_ID"] = 0;
		$comment_data["user_id"] = 0;
		$comment_data["comment_author"] = "";
		$comment_data["comment_author_email"] = "";
		$comment_data["comment_author_url"] = "";
	}

	$comment_id = wp_insert_comment($comment_data);
	if ($chatgpt_answer_type_of_users == "anonymously") {
		update_comment_meta($comment_id,'anonymously_user','anonymously');
	}
	add_comment_meta($comment_id,'askme_answer_chatgpt','chatgpt');
	update_post_meta($post_id,'askme_answer_chatgpt','chatgpt');
	add_comment_meta($comment_id,'comment_type','question');
	add_comment_meta($comment_id,'comment_vote',0);

	$mysql_time = current_time('mysql');

	if ($mysql_time != $time) {
		update_post_meta($post_id,'askme_answer_chatgpt_time','difference');
	}

	if ($mysql_time == $time && $chatgpt_best_answer == 1) {
		$the_best_answer = get_post_meta($post_id,"the_best_answer",true);
		if ($the_best_answer == "") {
			update_post_meta($post_id,"the_best_answer",$comment_id);
			$best_answer_user = get_user_meta($user_id,"best_answer_user",true);
			if (empty($best_answer_user)) {
				update_user_meta($user_id,"best_answer_user",array($comment_id));
			}else if (is_array($best_answer_user) && !in_array($comment_id,$best_answer_user)) {
				update_user_meta($user_id,"best_answer_user",array_merge($best_answer_user,array($comment_id)));
			}
			$the_best_answer = (int)get_user_meta($user_id,"the_best_answer",true);
			$the_best_answer++;
			update_user_meta($user_id,"the_best_answer",($the_best_answer < 0?0:$the_best_answer));
			$best_answer_option = (int)get_option("best_answer_option");
			$best_answer_option++;
			update_option("best_answer_option",($best_answer_option < 0?0:$best_answer_option));
			update_comment_meta($comment_id,"best_answer_comment","best_answer_comment");
		}
	}
}?>